import { decodePreviewableId } from '@previewjs/analyzer-api';
import { RPCs } from '@previewjs/api';
import { UNKNOWN_TYPE } from '@previewjs/type-analyzer';
import { createStackedReader, createFileSystemReader } from '@previewjs/vfs';
import express from 'express';
import fs$1 from 'fs-extra';
import { createRequire } from 'module';
import path from 'path';
import createLogger from 'pino';
import prettyLogger from 'pino-pretty';
import { exclusivePromiseRunner } from 'exclusive-promises';
import fs from 'fs';
import url from 'url';
import { globby } from 'globby';
import net from 'net';
import assertNever from 'assert-never';
import axios from 'axios';
import { createHttpTerminator } from 'http-terminator';
import viteTsconfigPaths from '@fwouts/vite-tsconfig-paths';
import { PREVIEW_CONFIG_NAME, readConfig } from '@previewjs/config';
import { polyfillNode } from 'esbuild-plugin-polyfill-node';
import fakeExportedTypesPlugin from 'rollup-plugin-friendly-type-imports';
import { loadTsconfig } from 'tsconfig-paths/lib/tsconfig-loader.js';
import * as vite from 'vite';
import { transformWithEsbuild, searchForWorkspaceRoot } from 'vite';
import { escape } from 'html-escaper';
import { Parser } from 'acorn';
import jsx from 'acorn-jsx';

function getCacheDir(rootDir) {
  try {
    const __dirname = url.fileURLToPath(new URL(".", import.meta.url));
    const { version } = JSON.parse(
      fs.readFileSync(path.resolve(__dirname, "..", "package.json"), "utf8")
    );
    return path.resolve(rootDir, "node_modules", ".previewjs", `v${version}`);
  } catch (e) {
    throw new Error(`Unable to detect @previewjs/core version.
${e}`);
  }
}

async function findFiles(rootDir, pattern, options = {}) {
  const gitRootPath = await findGitRoot(rootDir);
  const relativePath = path.relative(gitRootPath, rootDir);
  const relativePrefix = relativePath ? relativePath + path.sep : "";
  const files = await globby(relativePrefix + pattern, {
    gitignore: true,
    ignore: ["**/node_modules/**"],
    cwd: gitRootPath,
    absolute: true,
    followSymbolicLinks: false,
    suppressErrors: true,
    deep: options.maxDepth
  });
  let normalizedRootDirPath = rootDir.replace(/\\/g, "/");
  if (!normalizedRootDirPath.endsWith("/")) {
    normalizedRootDirPath += "/";
  }
  return files.filter((f) => f.startsWith(normalizedRootDirPath)).sort();
}
async function findGitRoot(dirPath, fallback = dirPath) {
  try {
    if (await fs$1.exists(path.join(dirPath, ".git"))) {
      return dirPath;
    } else {
      const parentDirPath = path.dirname(dirPath);
      if (!parentDirPath || parentDirPath === dirPath) {
        return fallback;
      }
      return findGitRoot(parentDirPath, fallback);
    }
  } catch {
    return fallback;
  }
}

const FILES_REQUIRING_REDETECTION = /* @__PURE__ */ new Set([
  "jsconfig.json",
  "tsconfig.json",
  "package.json",
  "package-lock.json",
  "pnpm-lock.yaml",
  "yarn.lock"
]);
const oneAtATime = exclusivePromiseRunner();
function crawlFiles(logger, workspace, frameworkPlugin, options = {}) {
  return oneAtATime(async () => {
    logger.debug(
      `Detecting components with options: ${JSON.stringify(options)}`
    );
    const detectionStartTimestamp = Date.now();
    const cacheFilePath = path.join(
      getCacheDir(workspace.rootDir),
      "components.json"
    );
    const absoluteFilePaths = await (async () => {
      if (options.filePaths) {
        return options.filePaths.map(
          (filePath) => path.join(workspace.rootDir, filePath)
        );
      } else {
        logger.debug(`Finding component files from root: ${workspace.rootDir}`);
        const filePaths = await findFiles(
          workspace.rootDir,
          "**/*.@(js|jsx|ts|tsx|svelte|vue)"
        );
        logger.debug(`Found ${filePaths.length} component files`);
        return filePaths;
      }
    })();
    const filePathsSet = new Set(
      absoluteFilePaths.map(
        (absoluteFilePath) => path.relative(workspace.rootDir, absoluteFilePath).replace(/\\/g, "/")
      )
    );
    let existingCache = {
      detectionStartTimestamp: 0,
      components: [],
      stories: []
    };
    if (fs$1.existsSync(cacheFilePath)) {
      try {
        existingCache = JSON.parse(
          fs$1.readFileSync(cacheFilePath, "utf8")
        );
      } catch (e) {
        logger.warn(`Unable to parse JSON from cache at ${cacheFilePath}`);
      }
    }
    if (existingCache.detectionStartTimestamp < await detectionMinimalTimestamp(workspace.rootDir)) {
      existingCache = {
        detectionStartTimestamp: 0,
        components: [],
        stories: []
      };
    }
    const changedAbsoluteFilePaths = absoluteFilePaths.filter(
      (absoluteFilePath) => {
        const entry = workspace.reader.readSync(absoluteFilePath);
        return entry?.kind === "file" && (options.forceRefresh || entry.lastModifiedMillis() >= existingCache.detectionStartTimestamp);
      }
    );
    const refreshedFilePaths = new Set(
      changedAbsoluteFilePaths.map(
        (absoluteFilePath) => path.relative(workspace.rootDir, absoluteFilePath).replace(/\\/g, "/")
      )
    );
    const shouldRecycle = ({ id }) => {
      const filePath = decodePreviewableId(id).filePath;
      return filePathsSet.has(filePath) && !refreshedFilePaths.has(filePath);
    };
    const recycledComponents = existingCache.components.filter(shouldRecycle);
    const recycledStories = existingCache.stories.filter(shouldRecycle);
    const { components: refreshedComponents, stories: refreshedStories } = await analyzeCore(
      logger,
      workspace,
      frameworkPlugin,
      changedAbsoluteFilePaths
    );
    const components = [...recycledComponents, ...refreshedComponents];
    const stories = [...recycledStories, ...refreshedStories];
    if (!options.filePaths) {
      await fs$1.mkdirp(path.dirname(cacheFilePath));
      const updatedCache = {
        detectionStartTimestamp,
        components,
        stories
      };
      await fs$1.writeFile(cacheFilePath, JSON.stringify(updatedCache));
    }
    return { components, stories };
  });
}
async function analyzeCore(logger, workspace, frameworkPlugin, changedAbsoluteFilePaths) {
  const components = [];
  const stories = [];
  if (changedAbsoluteFilePaths.length === 0) {
    return { components, stories };
  }
  logger.debug(
    `Running component detection with file paths:
- ${changedAbsoluteFilePaths.map(
      (absoluteFilePath) => path.relative(workspace.rootDir, absoluteFilePath)
    ).join("\n- ")}`
  );
  const found = await frameworkPlugin.crawlFiles(changedAbsoluteFilePaths);
  logger.debug(`Done running component detection`);
  for (const component of found.components) {
    components.push({
      id: component.id,
      sourcePosition: component.sourcePosition,
      exported: component.exported
    });
  }
  for (const story of found.stories) {
    stories.push({
      id: story.id,
      sourcePosition: story.sourcePosition,
      associatedComponentId: story.associatedComponent?.id || null
    });
  }
  return { components, stories };
}
async function detectionMinimalTimestamp(rootDir) {
  const nodeModulesPath = path.join(rootDir, "node_modules");
  let lastModifiedMillis = 0;
  if (await fs$1.pathExists(nodeModulesPath)) {
    for (const subdirectory of await fs$1.readdir(nodeModulesPath)) {
      const subdirectoryPath = path.join(nodeModulesPath, subdirectory);
      const stat = await fs$1.lstat(subdirectoryPath);
      if (stat.isDirectory() || stat.isSymbolicLink()) {
        lastModifiedMillis = Math.max(lastModifiedMillis, stat.mtimeMs);
      }
    }
  }
  for (const f of FILES_REQUIRING_REDETECTION) {
    const filePath = path.join(rootDir, f);
    if (await fs$1.pathExists(filePath)) {
      const stat = await fs$1.stat(filePath);
      lastModifiedMillis = Math.max(lastModifiedMillis, stat.mtimeMs);
    }
  }
  return lastModifiedMillis;
}

async function getFreePort(min) {
  for (let port = min; ; port++) {
    if (await isAvailable(port)) {
      return port;
    }
  }
}
async function isAvailable(port) {
  return new Promise((resolve) => {
    const server = net.createServer();
    server.unref();
    server.on("error", () => resolve(false));
    server.listen(port, () => {
      server.close(() => {
        resolve(true);
      });
    });
  });
}

const require$1 = createRequire(import.meta.url);
async function setupFrameworkPlugin({
  rootDir,
  frameworkPlugins,
  reader,
  logger
}) {
  const dependencies = await extractPackageDependencies(logger, rootDir);
  for (const candidate of frameworkPlugins) {
    if (await candidate.isCompatible(dependencies)) {
      return candidate.create({
        rootDir,
        reader,
        logger,
        dependencies
      });
    }
  }
  return null;
}
async function extractPackageDependencies(logger, rootDir) {
  const packageJsonPath = path.join(rootDir, "package.json");
  if (!await fs$1.pathExists(packageJsonPath)) {
    return {};
  }
  const { dependencies, devDependencies, peerDependencies } = JSON.parse(
    await fs$1.readFile(packageJsonPath, "utf8")
  );
  const allDependencies = {
    ...dependencies,
    ...devDependencies,
    ...peerDependencies
  };
  return Object.fromEntries(
    Object.entries(allDependencies).map(
      ([name, version]) => {
        let majorVersion;
        if (typeof version !== "string") {
          majorVersion = 0;
        } else if (version.startsWith("^") || version.startsWith("~")) {
          majorVersion = parseInt(version.slice(1));
        } else {
          majorVersion = parseInt(version);
        }
        const readInstalledVersion = async () => {
          try {
            const moduleEntryPath = findModuleEntryPath(name, rootDir);
            let packagePath = moduleEntryPath;
            let packageJsonPath2 = null;
            while (packagePath !== path.dirname(packagePath)) {
              const candidatePackageJsonPath = path.join(
                packagePath,
                "package.json"
              );
              if (fs$1.existsSync(candidatePackageJsonPath)) {
                packageJsonPath2 = candidatePackageJsonPath;
                break;
              }
              packagePath = path.dirname(packagePath);
            }
            if (!packageJsonPath2) {
              throw new Error(
                `No package.json path found from: ${moduleEntryPath}`
              );
            }
            const packageInfo = JSON.parse(
              await fs$1.readFile(packageJsonPath2, "utf8")
            );
            const version2 = packageInfo["version"];
            if (!version2 || typeof version2 !== "string") {
              throw new Error(
                `Invalid version found for package: ${packageJsonPath2}`
              );
            }
            return version2;
          } catch (e) {
            logger.error(
              `Unable to read installed version of package: ${name}`,
              e
            );
            return null;
          }
        };
        return [name, { majorVersion, readInstalledVersion }];
      }
    )
  );
}
function findModuleEntryPath(name, rootDir) {
  try {
    return require$1.resolve(name, {
      paths: [rootDir]
    });
  } catch (e) {
    if (e.code === "ERR_PACKAGE_PATH_NOT_EXPORTED") {
      const potentialMatch = e.message.match(
        /No "exports" main defined in (.*)(\/|\\)package\.json/
      );
      if (potentialMatch) {
        return potentialMatch[1];
      }
    }
    throw e;
  }
}

var __defProp$3 = Object.defineProperty;
var __defNormalProp$3 = (obj, key, value) => key in obj ? __defProp$3(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField$3 = (obj, key, value) => {
  __defNormalProp$3(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
var __accessCheck$1 = (obj, member, msg) => {
  if (!member.has(obj))
    throw TypeError("Cannot " + msg);
};
var __privateGet$1 = (obj, member, getter) => {
  __accessCheck$1(obj, member, "read from private field");
  return getter ? getter.call(obj) : member.get(obj);
};
var __privateAdd$1 = (obj, member, value) => {
  if (member.has(obj))
    throw TypeError("Cannot add the same private member more than once");
  member instanceof WeakSet ? member.add(obj) : member.set(obj, value);
};
var __privateSet$1 = (obj, member, value, setter) => {
  __accessCheck$1(obj, member, "write to private field");
  setter ? setter.call(obj, value) : member.set(obj, value);
  return value;
};
var _server;
class Server {
  constructor(options) {
    __publicField$3(this, "app");
    __publicField$3(this, "logger");
    __publicField$3(this, "serverTerminator", null);
    __privateAdd$1(this, _server, null);
    this.logger = options.logger;
    const app = express();
    app.use((_req, res, next) => {
      res.setHeader("Cache-Control", "max-age=0, must-revalidate");
      next();
    });
    for (const middleware of options.middlewares) {
      app.use(middleware);
    }
    this.app = app;
  }
  get server() {
    return __privateGet$1(this, _server);
  }
  async start(port) {
    return new Promise((resolve) => {
      const server = this.app.listen(port, () => {
        this.logger.info(
          `Preview.js server running at http://localhost:${port}.`
        );
        resolve(server);
      });
      this.serverTerminator = createHttpTerminator({
        server,
        gracefulTerminationTimeout: 0
      });
      __privateSet$1(this, _server, server);
    });
  }
  async stop() {
    if (this.serverTerminator) {
      await this.serverTerminator.terminate();
      this.serverTerminator = null;
    }
    __privateSet$1(this, _server, null);
    this.logger.info(`Preview.js server stopped.`);
  }
}
_server = new WeakMap();

function generateHtmlError(message) {
  return `<html>
  <head>
    <meta http-equiv="refresh" content="1">
    <style>
      body {
        background: #FCA5A5
      }
      pre {
        font-family: source-code-pro, Menlo, Monaco, Consolas, 'Courier New',
        monospace;
        font-size: 12px;
        line-height: 1.5em;
        color: #7F1D1D;
      }
    </style>
  </head>
  <body>
    <pre>${escape(message)}</pre>
  </body>
</html>`;
}

const extensions = /* @__PURE__ */ new Set([".js", ".jsx", ".ts", ".tsx"]);
const importRegexp = /import\s+([\S]+)\s+from\s+('|")([\S]+)\.(css|less|sass|scss|styl|stylus|postcss)(\?[\S]*)?&?('|")/g;
function cssModulesWithoutSuffixPlugin() {
  return {
    name: "previewjs:css-modules-without-suffix",
    async transform(code, id) {
      id = id.split("?")[0];
      if (!extensions.has(path.extname(id))) {
        return;
      }
      return code.replace(
        importRegexp,
        'import $1 from "$3.$4?$5&.module.css"'
      );
    }
  };
}

const jsExtensions$1 = /* @__PURE__ */ new Set([".js", ".jsx", ".ts", ".tsx"]);
function exportToplevelPlugin() {
  return {
    name: "previewjs:export-toplevel",
    enforce: "post",
    transform: async function(code, id) {
      if (id.indexOf(`/node_modules/`) !== -1) {
        return null;
      }
      id = id.split("?", 2)[0];
      const extension = path.extname(id);
      if (!jsExtensions$1.has(extension)) {
        return null;
      }
      try {
        const topLevelEntityNames = findTopLevelEntityNames(code);
        return `${code};
          export {
          ${topLevelEntityNames.map((c) => `${c} as __previewjs__${c},`).join("")}
        }`;
      } catch (e) {
        throw new Error(`Unable to parse ${id}: ${e}`);
      }
    }
  };
}
function findTopLevelEntityNames(source) {
  const parsed = Parser.extend(jsx()).parse(source, {
    ecmaVersion: "latest",
    sourceType: "module"
  });
  const topLevelEntityNames = [];
  for (const statement of parsed.body || []) {
    if (statement.type === "VariableDeclaration") {
      for (const declaration of statement.declarations) {
        if (declaration.type === "VariableDeclarator") {
          addIfIdentifier(topLevelEntityNames, declaration.id);
        }
      }
    }
    if (statement.type === "FunctionDeclaration") {
      addIfIdentifier(topLevelEntityNames, statement.id);
    }
    if (statement.type === "ClassDeclaration") {
      addIfIdentifier(topLevelEntityNames, statement.id);
    }
    if (statement.type === "ExportDefaultDeclaration") {
      if (statement.declaration?.id) {
        addIfIdentifier(topLevelEntityNames, statement.declaration.id);
      }
    }
    if (statement.type === "ExportNamedDeclaration") {
      if (statement.declaration?.id) {
        addIfIdentifier(topLevelEntityNames, statement.declaration.id);
      }
      if (statement.declaration?.declarations) {
        for (const declarator of statement.declaration.declarations) {
          addIfIdentifier(topLevelEntityNames, declarator.id);
        }
      }
    }
  }
  return [...new Set(topLevelEntityNames)];
}
function addIfIdentifier(array, id) {
  if (id && id.type === "Identifier") {
    const name = id.name;
    if (name.endsWith("_default")) {
      return;
    }
    array.push(name);
  }
}

function localEval() {
  return {
    name: "previewjs:local-eval",
    enforce: "post",
    transform: (code, id) => {
      if (id.indexOf(`/node_modules/`) !== -1) {
        return null;
      }
      if (id.indexOf("?") !== -1) {
        return null;
      }
      return code + // Note: This is an uppercase function so that it "looks like" a React component.
      // This needed to prevent breaking React Refresh, which expects all exports to
      // be React components.
      `
export const PreviewJsEvaluateLocally = async (autogenCallbackPropsSource, propsAssignmentSource, __jsxFactory__) => {
  let autogenCallbackProps = {};
  eval(autogenCallbackPropsSource);
  let properties = {};
  eval(propsAssignmentSource);
  return { autogenCallbackProps, properties };
}
`;
    }
  };
}

function publicAssetImportPluginPlugin({
  rootDir,
  publicDir
}) {
  return {
    name: "previewjs:public-asset-import",
    resolveId: async (source, importer) => {
      if (!importer || !source.startsWith("/") || source.startsWith("/__previewjs__/")) {
        return;
      }
      const publicDirAbsolutePath = path.join(rootDir, publicDir);
      const potentialPublicFilePath = path.join(publicDirAbsolutePath, source);
      if (!potentialPublicFilePath.startsWith(publicDirAbsolutePath + path.sep)) {
        return;
      }
      if (!await fs$1.pathExists(potentialPublicFilePath)) {
        return;
      }
      return source;
    }
  };
}

const VIRTUAL_PREFIX = `/@previewjs-virtual:`;
const VIRTUAL_PREFIX2 = `/@id/__x00__`;
const jsExtensions = /* @__PURE__ */ new Set([".js", ".jsx", ".ts", ".tsx"]);
function maybeVirtual(originalId) {
  let id = originalId;
  if (id.startsWith(VIRTUAL_PREFIX)) {
    id = id.slice(VIRTUAL_PREFIX.length);
  }
  if (id.startsWith("\0")) {
    id = id.slice(1);
  }
  if (id.startsWith(VIRTUAL_PREFIX2)) {
    id = id.slice(VIRTUAL_PREFIX2.length);
  }
  return [id, id !== originalId];
}
function virtualPlugin(options) {
  const { reader, rootDir } = options;
  return {
    name: "previewjs:virtual-fs",
    resolveId: async function(originalId, originalImporter) {
      const [importer, virtualImporter] = originalImporter ? maybeVirtual(originalImporter) : [null, false];
      if (originalId.indexOf(`/node_modules/`) !== -1) {
        return null;
      }
      let [id] = maybeVirtual(originalId);
      id = id.split("?", 2)[0];
      const extension = path.extname(id);
      let absoluteId;
      if (path.isAbsolute(id)) {
        absoluteId = id;
      } else {
        if (!importer) {
          return null;
        }
        if (extension && !jsExtensions.has(extension) && extension !== ".svg") {
          return path.join(path.dirname(importer), id);
        }
        absoluteId = path.join(path.dirname(importer), id);
      }
      const resolved = await resolveAbsoluteModuleId(absoluteId);
      if (resolved) {
        const [absoluteFilePath] = resolved;
        if (await fs$1.pathExists(absoluteFilePath) && !virtualImporter) {
          return null;
        }
        if (!absoluteId.endsWith(extension)) {
          absoluteId += extension;
        }
        const resolvedId = VIRTUAL_PREFIX + absoluteFilePath.replace(/\\/g, "/");
        return resolvedId;
      }
      return null;
    },
    load: async function(id) {
      if (id.indexOf(`/node_modules/`) !== -1) {
        return null;
      }
      if (id.startsWith(VIRTUAL_PREFIX)) {
        id = id.slice(VIRTUAL_PREFIX.length);
      }
      if (id.startsWith("\0")) {
        id = id.slice(1);
      }
      if (id.startsWith(VIRTUAL_PREFIX2)) {
        id = id.slice(VIRTUAL_PREFIX2.length);
      }
      const resolved = await resolveAbsoluteModuleId(id);
      if (!resolved) {
        return null;
      }
      const [absoluteFilePath, entry] = resolved;
      let isAllowed = false;
      for (const allowedAbsolutePath of options.allowedAbsolutePaths) {
        const relativePath = path.relative(
          allowedAbsolutePath,
          absoluteFilePath
        );
        if (relativePath && !relativePath.startsWith("..") && !path.isAbsolute(relativePath)) {
          isAllowed = true;
          break;
        }
      }
      if (!isAllowed) {
        options.logger.error(
          `Attempted access to ${absoluteFilePath} which is outside of allowed directories. See https://previewjs.com/docs/config and https://vitejs.dev/config/server-options.html#server-fs-allow for more information.`
        );
        return null;
      }
      if (entry.kind !== "file") {
        options.logger.error(`Unable to read file from ${absoluteFilePath}`);
        return null;
      }
      const source = await entry.read();
      const fileExtension = path.extname(absoluteFilePath);
      if (!jsExtensions.has(fileExtension)) {
        return source;
      }
      const moduleExtension = path.extname(id);
      return {
        code: (
          // We run an esbuild transform for .js or no extension
          // because it may include JSX. Otherwise, let plugins
          // decide whether to use esbuild or not.
          !id.includes("__previewjs_internal__") && (moduleExtension === "" || moduleExtension === ".js") ? (await transformWithEsbuild(source, absoluteFilePath, {
            loader: absoluteFilePath.endsWith(".ts") ? "ts" : "tsx",
            format: "esm",
            target: "es2020",
            sourcefile: path.relative(rootDir, absoluteFilePath),
            ...options.esbuildOptions
          })).code : source
        )
      };
    },
    transform: (code, id) => {
      if (!id.startsWith(VIRTUAL_PREFIX)) {
        return null;
      }
      return {
        code,
        map: {
          mappings: "",
          names: [],
          sources: [],
          version: 3
        }
      };
    },
    handleHotUpdate: async function(context) {
      const moduleGraph = options.moduleGraph();
      if (!moduleGraph) {
        return;
      }
      const filePath = path.relative(rootDir, context.file);
      const absoluteFilePath = path.join(rootDir, filePath);
      const entry = await reader.read(absoluteFilePath);
      if (!entry || entry.kind !== "file") {
        return;
      }
      const virtualModuleId = VIRTUAL_PREFIX + absoluteFilePath.replace(/\\/g, "/");
      const node = moduleGraph.getModuleById(virtualModuleId);
      return node && [node];
    }
  };
  async function resolveAbsoluteModuleId(id) {
    return await resolveBaseFilePath(id) || await resolveBaseFilePath(path.join(rootDir, id));
  }
  async function resolveBaseFilePath(baseFilePath) {
    for (const suffix of [
      "",
      ...jsExtensions,
      ...[...jsExtensions].map((extension) => `${path.sep}index${extension}`)
    ]) {
      const absoluteFilePath = `${baseFilePath}${suffix}`;
      const entry = await reader.read(absoluteFilePath);
      if (entry?.kind === "file") {
        return [absoluteFilePath, entry];
      }
    }
    return null;
  }
}

var __defProp$2 = Object.defineProperty;
var __defNormalProp$2 = (obj, key, value) => key in obj ? __defProp$2(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField$2 = (obj, key, value) => {
  __defNormalProp$2(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
var __accessCheck = (obj, member, msg) => {
  if (!member.has(obj))
    throw TypeError("Cannot " + msg);
};
var __privateGet = (obj, member, getter) => {
  __accessCheck(obj, member, "read from private field");
  return getter ? getter.call(obj) : member.get(obj);
};
var __privateAdd = (obj, member, value) => {
  if (member.has(obj))
    throw TypeError("Cannot add the same private member more than once");
  member instanceof WeakSet ? member.add(obj) : member.set(obj, value);
};
var __privateSet = (obj, member, value, setter) => {
  __accessCheck(obj, member, "write to private field");
  setter ? setter.call(obj, value) : member.set(obj, value);
  return value;
};
var _state, _exclusively, _readerListener;
const POSTCSS_CONFIG_FILE = [
  ".postcssrc",
  ".postcssrc.json",
  ".postcssrc.yml",
  ".postcssrc.js",
  ".postcssrc.mjs",
  ".postcssrc.cjs",
  ".postcssrc.ts",
  "postcss.config.js",
  "postcss.config.mjs",
  "postcss.config.cjs",
  "postcss.config.ts"
];
const GLOBAL_CSS_FILE_NAMES_WITHOUT_EXT = [
  "index",
  "global",
  "globals",
  "style",
  "styles",
  "app"
];
const GLOBAL_CSS_EXTS = ["css", "sass", "scss", "less", "styl", "stylus"];
const GLOBAL_CSS_FILE = GLOBAL_CSS_FILE_NAMES_WITHOUT_EXT.flatMap(
  (fileName) => GLOBAL_CSS_EXTS.map((ext) => `${fileName}.${ext}`)
);
const FILES_REQUIRING_VITE_RESTART = /* @__PURE__ */ new Set([
  PREVIEW_CONFIG_NAME,
  ...FILES_REQUIRING_REDETECTION,
  ...POSTCSS_CONFIG_FILE,
  ...GLOBAL_CSS_FILE,
  "vite.config.js",
  "vite.config.ts",
  // TODO: Make plugins contribute files requiring restart to make core agnostic of Svelte config files.
  "svelte.config.js"
]);
async function endState(state) {
  if (state?.kind === "starting") {
    return state.promise;
  } else {
    return state;
  }
}
class ViteManager {
  constructor(options) {
    this.options = options;
    __publicField$2(this, "middleware");
    __privateAdd(this, _state, null);
    __privateAdd(this, _exclusively, void 0);
    __privateAdd(this, _readerListener, {
      onChange: (absoluteFilePath, info) => {
        this.onFileChanged(path.resolve(absoluteFilePath), info).catch(
          (e) => this.options.logger.error(e)
        );
      }
    });
    const router = express.Router();
    router.use(async (req, res, next) => {
      const state = await endState(__privateGet(this, _state));
      if (!state || state.kind === "error") {
        return next();
      }
      try {
        state.viteServer.middlewares(req, res, next);
      } catch (e) {
        this.options.logger.error(`Vite middleware error: ${e}`);
        res.status(500).end(`Vite middleware error: ${e}`);
      }
    });
    this.middleware = router;
    __privateSet(this, _exclusively, exclusivePromiseRunner());
  }
  get state() {
    return __privateGet(this, _state);
  }
  async loadIndexHtml(url, id) {
    return __privateGet(this, _exclusively).call(this, async () => {
      const state = await endState(__privateGet(this, _state));
      if (!state) {
        return generateHtmlError(`Vite server is not running`);
      }
      if (state.kind === "error") {
        return generateHtmlError(state.error);
      }
      const template = await fs$1.readFile(
        this.options.shadowHtmlFilePath,
        "utf-8"
      );
      const { config, viteServer } = state;
      const { filePath, name: previewableName } = decodePreviewableId(id);
      const componentPath = filePath.replace(/\\/g, "/");
      const wrapper = config.wrapper;
      const wrapperPath = wrapper && await fs$1.pathExists(path.join(this.options.rootDir, wrapper.path)) ? wrapper.path.replace(/\\/g, "/") : null;
      return await viteServer.transformIndexHtml(
        url,
        template.replace(
          "<!-- %OPTIONAL_HEAD_CONTENT% -->",
          `
    <script type="module">
    import { initListeners, initPreview } from "/__previewjs_internal__/index.ts";

    initListeners();

    import.meta.hot.accept();

    let latestPreviewableModule;
    let latestWrapperModule;
    let refresh;

    import.meta.hot.accept(["/${componentPath}"], ([previewableModule]) => {
      if (previewableModule && refresh) {
        latestPreviewableModule = previewableModule;
        refresh(latestPreviewableModule, latestWrapperModule);
      }
    });

    ${wrapperPath ? `
    const wrapperModulePromise = import(/* @vite-ignore */ "/${wrapperPath}");
    import.meta.hot.accept(["/${wrapperPath}"], ([wrapperModule]) => {
      if (wrapperModule && refresh) {
        latestWrapperModule = wrapperModule;
        refresh(latestPreviewableModule, latestWrapperModule);
      }
    });
    ` : `
    const wrapperModulePromise = Promise.all([${config.detectedGlobalCssFilePaths.map(
            (cssFilePath) => `import(/* @vite-ignore */ "/${cssFilePath.replace(
              /\\/g,
              "/"
            )}").catch(() => null)`
          ).join(",")}]).then(() => null);
    `}

    // Important: the wrapper must be loaded first as it may monkey-patch
    // modules imported by the component module.
    wrapperModulePromise.then(wrapperModule => {
      latestWrapperModule = wrapperModule;
      import(/* @vite-ignore */ "/${componentPath}").then(previewableModule => {
        latestPreviewableModule = previewableModule;
        refresh = initPreview({
          previewableModule,
          previewableName: ${JSON.stringify(previewableName)},
          wrapperModule,
          wrapperName: ${JSON.stringify(wrapper?.componentName || null)},
        });
      });
    });
    <\/script>`
        )
      );
    });
  }
  // Note: this is guaranteed not to throw.
  async start() {
    let setEndState;
    __privateSet(this, _state, {
      kind: "starting",
      promise: new Promise((resolve) => {
        setEndState = (state) => {
          __privateSet(this, _state, state);
          resolve(state);
        };
      })
    });
    this.options.reader.listeners.add(__privateGet(this, _readerListener));
    try {
      process.chdir(this.options.rootDir);
      const configFromProject = await readConfig(this.options.rootDir);
      const globalCssAbsoluteFilePaths = await findFiles(
        this.options.rootDir,
        `**/@(${GLOBAL_CSS_FILE_NAMES_WITHOUT_EXT.join(
          "|"
        )}).@(${GLOBAL_CSS_EXTS.join("|")})`,
        {
          maxDepth: 3
        }
      );
      const config = {
        ...configFromProject,
        wrapper: configFromProject.wrapper || {
          path: this.options.frameworkPlugin.defaultWrapperPath
        },
        detectedGlobalCssFilePaths: globalCssAbsoluteFilePaths.map(
          (absoluteFilePath) => path.relative(this.options.rootDir, absoluteFilePath)
        )
      };
      const tsInferredAlias = [];
      let tsConfig = null;
      for (const potentialTsConfigFileName of [
        "tsconfig.json",
        "jsconfig.json"
      ]) {
        const potentialTsConfigFilePath = path.join(
          this.options.rootDir,
          potentialTsConfigFileName
        );
        if (await fs$1.pathExists(potentialTsConfigFilePath)) {
          tsConfig = loadTsconfig(potentialTsConfigFilePath) || null;
          if (tsConfig) {
            break;
          }
        }
      }
      this.options.logger.debug(
        `Loaded ts/jsconfig: ${JSON.stringify(tsConfig || null, null, 2)}`
      );
      const baseUrl = tsConfig?.compilerOptions?.baseUrl || "";
      const tsConfigPaths = tsConfig?.compilerOptions?.paths || {};
      let baseAlias = baseUrl.startsWith("./") ? baseUrl.substring(1) : baseUrl;
      if (baseAlias && !baseAlias.endsWith("/")) {
        baseAlias += "/";
      }
      for (const [match, mapping] of Object.entries(tsConfigPaths)) {
        const firstMapping = mapping[0];
        if (!firstMapping) {
          continue;
        }
        const matchNoWildcard = match.endsWith("/*") ? match.slice(0, match.length - 2) : match;
        const firstMappingNoWildcard = firstMapping.endsWith("/*") ? firstMapping.slice(0, firstMapping.length - 2) : firstMapping;
        tsInferredAlias.push({
          find: matchNoWildcard,
          replacement: path.join(
            this.options.rootDir,
            baseUrl,
            firstMappingNoWildcard
          )
        });
      }
      const existingViteConfig = await vite.loadConfigFromFile(
        {
          command: "serve",
          mode: "development"
        },
        void 0,
        this.options.rootDir
      );
      const defaultLogger = vite.createLogger(
        viteLogLevelFromPinoLogger(this.options.logger)
      );
      const frameworkPluginViteConfig = this.options.frameworkPlugin.viteConfig(
        await flattenPlugins([
          ...existingViteConfig?.config.plugins || [],
          ...config.vite?.plugins || []
        ])
      );
      const publicDir = config.vite?.publicDir || existingViteConfig?.config.publicDir || frameworkPluginViteConfig.publicDir || config.publicDir;
      const plugins = replaceHandleHotUpdate(
        this.options.reader,
        await flattenPlugins([
          viteTsconfigPaths({
            root: this.options.rootDir
          }),
          virtualPlugin({
            logger: this.options.logger,
            reader: this.options.reader,
            rootDir: this.options.rootDir,
            allowedAbsolutePaths: config.vite?.server?.fs?.allow || existingViteConfig?.config.server?.fs?.allow || [
              searchForWorkspaceRoot(this.options.rootDir)
            ],
            moduleGraph: () => {
              if (__privateGet(this, _state)?.kind === "running") {
                return __privateGet(this, _state).viteServer.moduleGraph;
              }
              return null;
            },
            esbuildOptions: frameworkPluginViteConfig.esbuild || {}
          }),
          localEval(),
          exportToplevelPlugin(),
          fakeExportedTypesPlugin({
            readFile: (absoluteFilePath) => this.options.reader.read(absoluteFilePath).then((entry) => {
              if (entry?.kind !== "file") {
                return null;
              }
              return entry.read();
            })
          }),
          cssModulesWithoutSuffixPlugin(),
          publicAssetImportPluginPlugin({
            rootDir: this.options.rootDir,
            publicDir
          }),
          frameworkPluginViteConfig.plugins
        ])
      );
      this.options.logger.debug(`Creating Vite server`);
      const viteServerPromise = vite.createServer({
        ...frameworkPluginViteConfig,
        ...existingViteConfig?.config,
        ...config.vite,
        configFile: false,
        root: this.options.rootDir,
        optimizeDeps: {
          entries: [],
          esbuildOptions: {
            // TODO: Remove this annotation once upgraded to Vite 5.
            // @ts-ignore incompatible esbuild versions with Vite 4
            plugins: [polyfillNode()]
          }
        },
        server: {
          middlewareMode: true,
          hmr: {
            overlay: false,
            server: this.options.server,
            clientPort: this.options.port,
            ...typeof config.vite?.server?.hmr === "object" ? config.vite?.server?.hmr : {}
          },
          fs: {
            strict: false,
            ...config.vite?.server?.fs || {}
          },
          ...config.vite?.server
        },
        customLogger: {
          info: defaultLogger.info,
          warn: defaultLogger.warn,
          error: (msg, options) => {
            if (!msg.startsWith("\x1B[31mInternal server error")) {
              if (__privateGet(this, _state)?.kind === "running") {
                __privateGet(this, _state).viteServer.ws.send({
                  type: "error",
                  err: {
                    message: msg,
                    stack: ""
                  }
                });
              }
            }
            defaultLogger.error(msg, options);
          },
          warnOnce: defaultLogger.warnOnce,
          clearScreen: () => {
          },
          hasWarned: defaultLogger.hasWarned,
          hasErrorLogged: defaultLogger.hasErrorLogged
        },
        clearScreen: false,
        cacheDir: config.vite?.cacheDir || existingViteConfig?.config.cacheDir || this.options.cacheDir,
        publicDir,
        plugins,
        define: {
          __filename: void 0,
          __dirname: void 0,
          ...frameworkPluginViteConfig.define,
          ...existingViteConfig?.config.define,
          ...config.vite?.define
        },
        resolve: {
          ...existingViteConfig?.config.resolve,
          ...config.vite?.resolve,
          alias: [
            // First defined rules are applied first, therefore highest priority should come first.
            ...viteAliasToRollupAliasEntries(config.vite?.resolve?.alias),
            ...viteAliasToRollupAliasEntries(config.alias),
            ...viteAliasToRollupAliasEntries(
              existingViteConfig?.config.resolve?.alias
            ),
            ...tsInferredAlias,
            {
              find: /^~(.*)/,
              replacement: baseAlias + "$1"
            },
            {
              find: "@",
              replacement: baseAlias
            },
            ...viteAliasToRollupAliasEntries(
              frameworkPluginViteConfig.resolve?.alias
            )
          ]
        }
      });
      const viteServer = await viteServerPromise;
      setEndState({
        kind: "running",
        config,
        viteServer
      });
      this.options.logger.debug(`Done starting Vite server`);
    } catch (e) {
      this.options.logger.error(`Vite startup error: ${e}`);
      setEndState({
        kind: "error",
        error: e.stack || e.message
      });
    }
    return __privateGet(this, _state).promise;
  }
  async stop({ restart } = { restart: false }) {
    await __privateGet(this, _exclusively).call(this, async () => {
      this.options.reader.listeners.remove(__privateGet(this, _readerListener));
      const state = await endState(__privateGet(this, _state));
      if (state?.kind === "running") {
        await state.viteServer.close();
      }
      if (restart) {
        return this.start();
      } else {
        return __privateSet(this, _state, null);
      }
    });
  }
  async onFileChanged(absoluteFilePath, info) {
    const state = await endState(this.state);
    if (!state) {
      return;
    }
    if (!info.virtual) {
      this.options.frameworkPlugin.typeAnalyzer.invalidateCachedTypesForFile(
        path.relative(this.options.rootDir, absoluteFilePath)
      );
      if (FILES_REQUIRING_VITE_RESTART.has(path.basename(absoluteFilePath))) {
        this.options.logger.info(
          "New dependencies were detected. Restarting..."
        );
        await this.stop({ restart: true });
        return;
      }
    }
    if (state.kind !== "running") {
      return;
    }
    const { viteServer, config } = state;
    if (info.virtual) {
      const modules = await viteServer.moduleGraph.getModulesByFile(
        absoluteFilePath
      );
      for (const module of modules || []) {
        if (!module.id) {
          continue;
        }
        try {
          const loaded = await viteServer.pluginContainer.load(module.id);
          if (!loaded) {
            continue;
          }
          const source = typeof loaded === "object" ? loaded.code : loaded;
          await viteServer.pluginContainer.transform(source, module.id);
        } catch (e) {
          return;
        }
      }
      for (const onChange of viteServer.watcher.listeners("change")) {
        onChange(absoluteFilePath);
      }
    } else if (config.wrapper && absoluteFilePath === path.resolve(this.options.rootDir, config.wrapper.path)) {
      viteServer.moduleGraph.invalidateAll();
      viteServer.ws.send({
        type: "full-reload"
      });
    }
  }
}
_state = new WeakMap();
_exclusively = new WeakMap();
_readerListener = new WeakMap();
function viteAliasToRollupAliasEntries(alias) {
  if (!alias) {
    return [];
  }
  if (Array.isArray(alias)) {
    return alias;
  } else {
    return Object.entries(alias).map(([find, replacement]) => ({
      find,
      replacement
    }));
  }
}
async function flattenPlugins(pluginOptions) {
  const plugins = [];
  for (const pluginOption of await Promise.all(pluginOptions)) {
    if (!pluginOption) {
      continue;
    }
    if (Array.isArray(pluginOption)) {
      plugins.push(...await flattenPlugins(pluginOption));
    } else {
      plugins.push(pluginOption);
    }
  }
  return plugins;
}
function viteLogLevelFromPinoLogger(logger) {
  switch (logger.level) {
    case "fatal":
      return "silent";
    case "error":
      return "error";
    case "warn":
      return "warn";
    case "info":
      return "info";
    case "debug":
      return "info";
    case "trace":
      return "info";
    case "silent":
      return "silent";
    default:
      logger.warn(`Unknown log level: ${logger.level}`);
      return "info";
  }
}
function replaceHandleHotUpdate(reader, plugins) {
  return plugins.map(async (plugin) => {
    if (!plugin.handleHotUpdate) {
      return plugin;
    }
    const handleHotUpdate = typeof plugin.handleHotUpdate === "function" ? plugin.handleHotUpdate : plugin.handleHotUpdate.handler;
    return {
      ...plugin,
      handleHotUpdate: async (ctx) => {
        await handleHotUpdate({
          ...ctx,
          read: async () => {
            const entry = await reader.read(ctx.file);
            if (entry?.kind !== "file") {
              return ctx.read();
            }
            return entry.read();
          }
        });
      }
    };
  });
}

var __defProp$1 = Object.defineProperty;
var __defNormalProp$1 = (obj, key, value) => key in obj ? __defProp$1(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField$1 = (obj, key, value) => {
  __defNormalProp$1(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
class Previewer {
  constructor(options) {
    this.options = options;
    __publicField$1(this, "transformingReader");
    __publicField$1(this, "appServer", null);
    __publicField$1(this, "viteManager", null);
    __publicField$1(this, "status", { kind: "stopped" });
    __publicField$1(this, "disposeObserver", null);
    this.transformingReader = createStackedReader([
      options.reader,
      createFileSystemReader({
        mapping: {
          from: options.frameworkPlugin.previewDirPath,
          to: path.join(options.rootDir, "__previewjs_internal__", "renderer")
        },
        watch: false
      }),
      createFileSystemReader({
        mapping: {
          from: options.previewDirPath,
          to: options.rootDir
        },
        watch: false
      })
    ]);
  }
  async start(options = {}) {
    const statusBeforeStart = this.status;
    switch (statusBeforeStart.kind) {
      case "starting":
        try {
          await statusBeforeStart.promise;
        } catch (e) {
          this.options.logger.error(e);
          this.status = {
            kind: "stopped"
          };
          await this.startFromStopped();
        }
        break;
      case "started":
        break;
      case "stopping":
        try {
          await statusBeforeStart.promise;
        } catch (e) {
          this.options.logger.error(e);
          this.status = {
            kind: "stopped"
          };
        }
        await this.start(options);
        break;
      case "stopped":
        await this.startFromStopped();
        break;
      default:
        throw assertNever(statusBeforeStart);
    }
  }
  async startFromStopped() {
    this.status = {
      kind: "starting",
      promise: (async () => {
        const router = express.Router();
        router.get(/^\/.*:[^/]+\/$/, async (req, res, next) => {
          if (req.url.includes("?html-proxy")) {
            next();
            return;
          }
          const previewableId = req.path.substring(1, req.path.length - 1);
          if (req.header("Accept") === "text/x-vite-ping") {
            res.writeHead(204).end();
            return;
          }
          if (!this.viteManager) {
            res.status(404).end(`Uh-Oh! Vite server is not running.`);
            return;
          }
          res.status(200).set({ "Content-Type": "text/html" }).end(
            await this.viteManager.loadIndexHtml(
              req.originalUrl,
              previewableId
            )
          );
        });
        router.use("/ping", async (req, res) => {
          res.json(
            JSON.stringify({
              pong: "match!"
            })
          );
        });
        this.appServer = new Server({
          logger: this.options.logger,
          middlewares: [
            ...this.options.middlewares || [],
            router,
            (req, res, next) => {
              this.viteManager?.middleware(req, res, next);
            }
          ]
        });
        if (this.transformingReader.observe) {
          this.disposeObserver = await this.transformingReader.observe(
            this.options.rootDir
          );
        }
        this.options.logger.debug(`Starting server`);
        const server = await this.appServer.start(this.options.port);
        this.options.logger.debug(`Starting Vite manager`);
        this.viteManager = new ViteManager({
          rootDir: this.options.rootDir,
          shadowHtmlFilePath: path.join(
            this.options.previewDirPath,
            "index.html"
          ),
          reader: this.transformingReader,
          cacheDir: path.join(getCacheDir(this.options.rootDir), "vite"),
          logger: this.options.logger,
          frameworkPlugin: this.options.frameworkPlugin,
          server,
          port: this.options.port
        });
        this.viteManager.start();
        this.options.logger.debug(`Previewer ready`);
        this.status = {
          kind: "started"
        };
      })()
    };
    await this.status.promise;
    for (let i = 0; ; i++) {
      try {
        await axios.get(`http://localhost:${this.options.port}`);
        break;
      } catch (e) {
        if (i === 10) {
          throw e;
        } else {
          await new Promise((resolve) => setTimeout(resolve, 50));
        }
      }
    }
  }
  async stop() {
    if (this.status.kind === "starting") {
      try {
        await this.status.promise;
      } catch {
        this.status = {
          kind: "stopped"
        };
      }
    }
    if (this.status.kind !== "started") {
      return;
    }
    this.status = {
      kind: "stopping",
      promise: (async () => {
        await this.disposeObserver?.();
        this.disposeObserver = null;
        await this.viteManager?.stop();
        this.viteManager = null;
        await this.appServer?.stop();
        this.appServer = null;
        this.status = {
          kind: "stopped"
        };
      })()
    };
    await this.status.promise;
  }
}

var __defProp = Object.defineProperty;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField = (obj, key, value) => {
  __defNormalProp(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
class ApiRouter {
  constructor(logger) {
    this.logger = logger;
    __publicField(this, "handlers", /* @__PURE__ */ new Map());
    __publicField(this, "registerRPC", (endpoint, handler) => {
      if (this.handlers.has(endpoint.path)) {
        throw new Error(
          `Multiple handlers registered for endpoint '${endpoint.path}'`
        );
      }
      this.handlers.set(endpoint.path, handler);
    });
  }
  async handle(path, request) {
    const handler = this.handlers.get(path);
    if (!handler) {
      return {
        kind: "error",
        message: `No handler is registered for path '${path}'`
      };
    }
    this.logger.debug(`RPC ${path} invoked`);
    try {
      const response = await handler(request);
      return {
        kind: "success",
        response
      };
    } catch (e) {
      this.logger.error(`RPC ${path} failed`, e);
      return {
        kind: "error",
        message: e.message
      };
    }
  }
}

const require = createRequire(import.meta.url);
process.on("uncaughtException", (e) => {
  console.error("Encountered an uncaught exception", e);
});
process.on("unhandledRejection", (e) => {
  console.error("Encountered an unhandled promise", e);
});
class NoCompatiblePluginError extends Error {
  constructor(message) {
    super(message);
  }
}
async function createWorkspace({
  rootDir,
  frameworkPlugins,
  logger = createLogger(
    { level: process.env["PREVIEWJS_LOG_LEVEL"]?.toLowerCase() || "warn" },
    prettyLogger({ colorize: true })
  ),
  reader = createFileSystemReader(),
  onServerStart = () => Promise.resolve({})
}) {
  const frameworkPlugin = await setupFrameworkPlugin({
    rootDir,
    frameworkPlugins,
    reader,
    logger
  });
  if (!frameworkPlugin) {
    throw new NoCompatiblePluginError(
      `No compatible plugin found for workspace with root: ${rootDir}`
    );
  }
  logger.debug(
    `Creating workspace with framework plugin ${frameworkPlugin.name} from root: ${rootDir}`
  );
  const expectedPluginApiVersion = 4;
  if (!frameworkPlugin.pluginApiVersion || frameworkPlugin.pluginApiVersion < expectedPluginApiVersion) {
    throw new Error(
      `Preview.js framework plugin ${frameworkPlugin.name} is incompatible with this version of Preview.js. Please upgrade it.`
    );
  } else if (frameworkPlugin.pluginApiVersion > expectedPluginApiVersion) {
    throw new Error(
      `Preview.js framework plugin ${frameworkPlugin.name} is too recent. Please upgrade Preview.js or use an older version of ${frameworkPlugin.name}.`
    );
  }
  const activePreviewers = /* @__PURE__ */ new Set();
  const workspace = {
    frameworkPluginName: frameworkPlugin.name,
    crawlFiles: frameworkPlugin.crawlFiles,
    typeAnalyzer: frameworkPlugin.typeAnalyzer,
    rootDir,
    reader,
    startServer: async ({ port } = {}) => {
      port || (port = await getFreePort(3140));
      const router = new ApiRouter(logger);
      router.registerRPC(RPCs.Analyze, async ({ previewableIds }) => {
        logger.debug(`Analyzing: ${previewableIds.join(", ")}`);
        const detected = await frameworkPlugin.crawlFiles([
          ...new Set(
            previewableIds.map(
              (c) => path.join(rootDir, decodePreviewableId(c).filePath)
            )
          )
        ]);
        logger.debug(
          `Detected ${detected.components.length} components and ${detected.stories.length} stories`
        );
        const previewables = [];
        const idToDetectedComponent = Object.fromEntries(
          detected.components.map((c) => [c.id, c])
        );
        const idToDetectedStory = Object.fromEntries(
          detected.stories.map((c) => [c.id, c])
        );
        let types = {};
        for (const id of previewableIds) {
          const component = idToDetectedComponent[id];
          const story = idToDetectedStory[id];
          let componentTypes;
          if (component) {
            logger.debug(`Analyzing component: ${id}`);
            let props;
            ({ props, types: componentTypes } = await component.analyze());
            logger.debug(`Done analyzing: ${id}`);
            previewables.push({
              kind: "component",
              id,
              exported: component.exported,
              sourcePosition: component.sourcePosition,
              props
            });
          } else if (story) {
            let props;
            if (story.associatedComponent) {
              logger.debug(`Analyzing story: ${id}`);
              ({ props, types: componentTypes } = await story.associatedComponent.analyze());
              logger.debug(`Done analyzing: ${id}`);
            } else {
              logger.debug(`No associated component for story: ${id}`);
              props = UNKNOWN_TYPE;
              componentTypes = {};
            }
            previewables.push({
              kind: "story",
              id,
              sourcePosition: story.sourcePosition,
              associatedComponentId: story.associatedComponent?.id || null,
              props,
              args: (await story.analyze()).args
            });
          } else {
            const { filePath, name } = decodePreviewableId(id);
            throw new Error(`Component ${name} not detected in ${filePath}.`);
          }
          types = { ...types, ...componentTypes };
        }
        return {
          previewables,
          types
        };
      });
      router.registerRPC(
        RPCs.CrawlFiles,
        (options) => crawlFiles(logger, workspace, frameworkPlugin, options)
      );
      const middlewares = [
        express.json(),
        async (req, res, next) => {
          if (req.path.startsWith("/api/")) {
            res.json(await router.handle(req.path.substr(5), req.body));
          } else {
            next();
          }
        }
      ];
      if (onServerStart) {
        const { middlewares: additionalMiddlewares = [] } = await onServerStart(
          {
            registerRPC: (endpoint, handler) => router.registerRPC(endpoint, handler),
            workspace
          }
        );
        middlewares.push(...additionalMiddlewares);
      }
      const previewer = new Previewer({
        reader,
        rootDir,
        // TODO: Use a cleaner approach.
        previewDirPath: path.join(
          path.dirname(path.dirname(require.resolve("@previewjs/iframe"))),
          "preview"
        ),
        frameworkPlugin,
        logger,
        middlewares,
        port
      });
      await previewer.start();
      activePreviewers.add(previewer);
      return {
        url: () => `http://localhost:${port}`,
        stop: async () => {
          activePreviewers.delete(previewer);
          await previewer.stop();
        }
      };
    },
    dispose: async () => {
      await Promise.all(
        [...activePreviewers].map((previewer) => previewer.stop())
      );
      activePreviewers.clear();
      frameworkPlugin.dispose();
    }
  };
  return workspace;
}
function findWorkspaceRoot(absoluteFilePath) {
  let dirPath = path.resolve(absoluteFilePath);
  while (dirPath !== path.dirname(dirPath)) {
    if (fs$1.existsSync(path.join(dirPath, "package.json"))) {
      return dirPath;
    }
    dirPath = path.dirname(dirPath);
  }
  return null;
}

export { NoCompatiblePluginError, createWorkspace, findWorkspaceRoot };
